from typing import Dict, List
from mlserver.types import MetadataTensor

METADATA: Dict[str, Dict[str, List[MetadataTensor]]] = {
    "audio-classification": dict(
        inputs=[
            # file path inputs
            MetadataTensor(
                name="inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
            # file content bytes inputs
            MetadataTensor(
                name="inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="base64"),
            ),
            # numpy.ndarray inputs
            MetadataTensor(
                name="inputs",
                shape=[-1],
                datatype="FP32",
                parameters=dict(content_type="nplist"),
            ),
        ],
        outputs=[
            MetadataTensor(
                name="output_*",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="hg_json"),
            ),
        ],
    ),
    "automatic-speech-recognition": dict(
        inputs=[
            # file path inputs
            MetadataTensor(
                name="inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
            # file content bytes inputs
            MetadataTensor(
                name="inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="base64"),
            ),
            # numpy.ndarray inputs
            MetadataTensor(
                name="inputs",
                shape=[-1],
                datatype="FP32",
                parameters=dict(content_type="nplist"),
            ),
        ],
        outputs=[
            MetadataTensor(
                name="output_*",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="hg_json"),
            ),
        ],
    ),
    "feature-extraction": dict(
        inputs=[
            MetadataTensor(
                name="args",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
        ],
        outputs=[
            MetadataTensor(
                name="outputs",
                shape=[-1],
                datatype="FP64",
                parameters=dict(content_type="pd"),
            ),
        ],
    ),
    "text-classification": dict(
        inputs=[
            MetadataTensor(
                name="args",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
        ],
        outputs=[
            MetadataTensor(
                name="outputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="hg_json"),
            ),
        ],
    ),
    "token-classification": dict(
        inputs=[
            MetadataTensor(
                name="args",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
        ],
        outputs=[],
    ),
    "question-answering": dict(
        inputs=[
            MetadataTensor(
                name="context",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
            MetadataTensor(
                name="question",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
        ],
        outputs=[],
    ),
    "table-question-answering": dict(
        inputs=[
            MetadataTensor(
                name="array_inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="hg_jsonlist"),
            ),
        ],
        outputs=[],
    ),
    "visual-question-answering": dict(
        inputs=[
            MetadataTensor(
                name="array_inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="hg_jsonlist"),
            ),
        ],
        outputs=[],
    ),
    "fill-mask": dict(
        inputs=[
            MetadataTensor(
                name="array_inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
        ],
        outputs=[],
    ),
    "summarization": dict(
        inputs=[
            MetadataTensor(
                name="array_inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
        ],
        outputs=[],
    ),
    "translation": dict(
        inputs=[
            MetadataTensor(
                name="array_inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
        ],
        outputs=[],
    ),
    "text2text-generation": dict(
        inputs=[
            MetadataTensor(
                name="array_inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
        ],
        outputs=[],
    ),
    "text-generation": dict(
        inputs=[
            MetadataTensor(
                name="array_inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
        ],
        outputs=[],
    ),
    "zero-shot-classification": dict(
        inputs=[
            MetadataTensor(
                name="array_inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
            MetadataTensor(
                name="candidate_labels",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
        ],
        outputs=[],
    ),
    "zero-shot-image-classification": dict(
        inputs=[
            # file path inputs
            MetadataTensor(
                name="array_inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
            # file content bytes inputs
            MetadataTensor(
                name="array_inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="base64"),
            ),
            # numpy.ndarray inputs
            MetadataTensor(
                name="array_inputs",
                shape=[-1],
                datatype="FP32",
                parameters=dict(content_type="nplist"),
            ),
            MetadataTensor(
                name="candidate_labels",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
        ],
        outputs=[],
    ),
    "conversational": dict(
        inputs=[
            MetadataTensor(
                name="array_inputs",
                shape=[-1],
                datatype="FP32",
                parameters=dict(content_type="hg_conversation"),
            ),
        ],
        outputs=[],
    ),
    "image-classification": dict(
        inputs=[
            # file path inputs
            MetadataTensor(
                name="inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
            # file content bytes inputs
            MetadataTensor(
                name="inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="base64"),
            ),
            # pillow image  inputs
            MetadataTensor(
                name="inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="pillow_image"),
            ),
        ],
        outputs=[],
    ),
    "image-segmentation": dict(
        inputs=[
            # file path inputs
            MetadataTensor(
                name="inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
            # file content bytes inputs
            MetadataTensor(
                name="inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="base64"),
            ),
            # pillow image  inputs
            MetadataTensor(
                name="inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="pillow_image"),
            ),
        ],
        outputs=[],
    ),
    "object-detection": dict(
        inputs=[
            # file path inputs
            MetadataTensor(
                name="inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="str"),
            ),
            # file content bytes inputs
            MetadataTensor(
                name="inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="base64"),
            ),
            # pillow image  inputs
            MetadataTensor(
                name="inputs",
                shape=[-1],
                datatype="BYTES",
                parameters=dict(content_type="pillow_image"),
            ),
        ],
        outputs=[],
    ),
}



import asyncio

from mlserver.model import MLModel
from mlserver.settings import ModelSettings
from mlserver.logging import logger
from mlserver.types import (
    InferenceRequest,
    InferenceResponse,
)

from .settings import get_huggingface_settings
from .common import load_pipeline_from_settings
from .codecs import HuggingfaceRequestCodec
from .metadata import METADATA


class HuggingFaceRuntime(MLModel):
    """Runtime class for specific Huggingface models"""

    def __init__(self, settings: ModelSettings):
        self.hf_settings = get_huggingface_settings(settings)
        super().__init__(settings)

    async def load(self) -> bool:
        # Loading & caching pipeline in asyncio loop to avoid blocking
        logger.info(f"Loading model for task '{self.hf_settings.task_name}'...")
        await asyncio.get_running_loop().run_in_executor(
            None,
            load_pipeline_from_settings,
            self.hf_settings,
            self.settings,
        )

        # Now we load the cached model which should not block asyncio
        self._model = load_pipeline_from_settings(self.hf_settings, self.settings)
        self._merge_metadata()
        return True

    async def predict(self, payload: InferenceRequest) -> InferenceResponse:
        # TODO: convert and validate?
        kwargs = self.decode_request(payload, default_codec=HuggingfaceRequestCodec)
        args = kwargs.pop("args", [])

        array_inputs = kwargs.pop("array_inputs", [])
        if array_inputs:
            args = [list(array_inputs)] + args
        prediction = self._model(*args, **kwargs)

        return self.encode_response(
            payload=prediction, default_codec=HuggingfaceRequestCodec
        )

    def _merge_metadata(self) -> None:
        meta = METADATA.get(self.hf_settings.task)
        if meta:
            self.inputs += meta.get("inputs", [])  # type: ignore
            self.outputs += meta.get("outputs", [])  # type: ignore



from typing import List, Any
from mlserver.codecs.base import InputCodec, register_input_codec
from mlserver.types import RequestInput, ResponseOutput, Parameters
from transformers.pipelines import Conversation
from mlserver.codecs.lists import is_list_of
from .utils import json_decode, json_encode


@register_input_codec
class HuggingfaceConversationCodec(InputCodec):
    """
    Codec that convers to / from a transformers Conversation input.
    """

    ContentType = "hg_conversation"

    @classmethod
    def can_encode(cls, payload: Any) -> bool:
        return is_list_of(payload, Conversation)

    @classmethod
    def encode_output(
        cls, name: str, payload: List[Conversation], use_bytes: bool = True, **kwargs
    ) -> ResponseOutput:
        encoded = [json_encode(item, use_bytes=use_bytes) for item in payload]
        shape = [len(encoded), 1]
        return ResponseOutput(
            name=name,
            parameters=Parameters(
                content_type=cls.ContentType,
            ),
            datatype="BYTES",
            shape=shape,
            data=encoded,
        )

    @classmethod
    def decode_output(cls, response_output: ResponseOutput) -> List[Any]:
        packed = response_output.data.__root__
        return [json_decode(item) for item in packed]

    @classmethod
    def encode_input(
        cls, name: str, payload: List[Conversation], use_bytes: bool = True, **kwargs
    ) -> RequestInput:
        output = cls.encode_output(name, payload, use_bytes)
        return RequestInput(
            name=output.name,
            parameters=Parameters(
                content_type=cls.ContentType,
            ),
            datatype=output.datatype,
            shape=output.shape,
            data=output.data,
        )

    @classmethod
    def decode_input(cls, request_input: RequestInput) -> List[Conversation]:
        packed = request_input.data.__root__
        return [json_decode(item) for item in packed]



import json
import numpy as np

from typing import Callable
from functools import partial
from mlserver.settings import ModelSettings

import torch
import tensorflow as tf

from optimum.pipelines import pipeline as opt_pipeline
from transformers.pipelines import pipeline as trf_pipeline
from transformers.pipelines.base import Pipeline

from .settings import HuggingFaceSettings


OPTIMUM_ACCELERATOR = "ort"

_PipelineConstructor = Callable[..., Pipeline]


def load_pipeline_from_settings(
    hf_settings: HuggingFaceSettings, settings: ModelSettings
) -> Pipeline:
    # TODO: Support URI for locally downloaded artifacts
    # uri = model_parameters.uri
    pipeline = _get_pipeline_class(hf_settings)

    batch_size = 1
    if settings.max_batch_size:
        batch_size = settings.max_batch_size

    tokenizer = hf_settings.pretrained_tokenizer
    if not tokenizer:
        tokenizer = hf_settings.pretrained_model
    if hf_settings.framework == "tf":
        if hf_settings.inter_op_threads is not None:
            tf.config.threading.set_inter_op_parallelism_threads(
                hf_settings.inter_op_threads
            )
        if hf_settings.intra_op_threads is not None:
            tf.config.threading.set_intra_op_parallelism_threads(
                hf_settings.intra_op_threads
            )
    elif hf_settings.framework == "pt":
        if hf_settings.inter_op_threads is not None:
            torch.set_num_interop_threads(hf_settings.inter_op_threads)
        if hf_settings.intra_op_threads is not None:
            torch.set_num_threads(hf_settings.intra_op_threads)

    hf_pipeline = pipeline(
        hf_settings.task_name,
        model=hf_settings.pretrained_model,
        tokenizer=tokenizer,
        device=hf_settings.device,
        batch_size=batch_size,
        framework=hf_settings.framework,
    )

    # If max_batch_size > 0 we need to ensure tokens are padded
    if settings.max_batch_size:
        model = hf_pipeline.model
        eos_token_id = model.config.eos_token_id
        hf_pipeline.tokenizer.pad_token_id = [str(eos_token_id)]  # type: ignore

    return hf_pipeline


def _get_pipeline_class(hf_settings: HuggingFaceSettings) -> _PipelineConstructor:
    if hf_settings.optimum_model:
        return partial(opt_pipeline, accelerator=OPTIMUM_ACCELERATOR)

    return trf_pipeline


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)



import numpy as np

from typing import Any, List

from mlserver.types import RequestInput, ResponseOutput, Parameters

from mlserver.codecs.base import InputCodec, register_input_codec
from mlserver.codecs.utils import inject_batch_dimension
from mlserver.codecs.lists import is_list_of
from mlserver.codecs.numpy import to_datatype, _to_ndarray, _encode_data


@register_input_codec
class NumpyListCodec(InputCodec):
    """
    Encodes a tensor as a numpy array list.
    """

    ContentType = "nplist"

    @classmethod
    def can_encode(csl, payload: Any) -> bool:
        if not is_list_of(payload, np.ndarray):
            return False
        # only the support same shaped ndarray
        return len(set([matrix.shape for matrix in payload])) == 1

    @classmethod
    def encode_output(
        cls, name: str, payload: List[np.ndarray], **kwargs
    ) -> ResponseOutput:
        # NOTICE: composed np.array may cause loss of accuracy
        composed = np.array(payload)
        datatype = to_datatype(composed.dtype)
        shape = inject_batch_dimension(list(composed.shape))

        return ResponseOutput(
            name=name,
            datatype=datatype,
            shape=shape,
            data=_encode_data(composed, datatype),
            parameters=Parameters(content_type=cls.ContentType),
        )

    @classmethod
    def decode_output(cls, response_output: ResponseOutput) -> List[np.ndarray]:
        return cls.decode_input(response_output)  # type: ignore

    @classmethod
    def encode_input(
        cls, name: str, payload: List[np.ndarray], **kwargs
    ) -> RequestInput:
        output = cls.encode_output(name=name, payload=payload)

        return RequestInput(
            name=output.name,
            datatype=output.datatype,
            shape=output.shape,
            data=output.data,
            parameters=Parameters(content_type=cls.ContentType),
        )

    @classmethod
    def decode_input(cls, request_input: RequestInput) -> List[np.ndarray]:
        model_data = _to_ndarray(request_input)
        reshaped = model_data.reshape(request_input.shape)
        return [el for el in reshaped]



from typing import Any, Union
from mlserver.types import RequestInput, ResponseOutput, Parameters
from mlserver.codecs.base import InputCodec, register_input_codec


@register_input_codec
class RawCodec(InputCodec):
    """
    Encode/Decode raw python datatypes
    """

    ContentType = "raw"

    @classmethod
    def can_encode(cls, payload: Any) -> bool:
        return (
            isinstance(payload, int)
            or isinstance(payload, str)
            or isinstance(payload, float)
        )

    @classmethod
    def encode_output(
        cls, name: str, payload: Union[int, str, float], **kwargs
    ) -> ResponseOutput:
        return ResponseOutput(
            name=name,
            datatype="BYTES",
            shape=[1],
            data=[payload],
            parameters=Parameters(content_type=cls.ContentType),
        )

    @classmethod
    def decode_output(cls, response_output: ResponseOutput) -> Union[int, str, float]:
        return cls.decode_input(response_output)  # type: ignore

    @classmethod
    def encode_input(
        cls, name: str, payload: Union[int, str, float], **kwargs
    ) -> RequestInput:
        output = cls.encode_output(name=name, payload=payload)

        return RequestInput(
            name=output.name,
            datatype=output.datatype,
            shape=output.shape,
            data=output.data,
            parameters=output.parameters,
        )

    @classmethod
    def decode_input(cls, request_input: RequestInput) -> Union[int, str, float]:
        return request_input.data[0]



import io
import base64
from typing import List, Any, Union
from PIL import Image
from mlserver.codecs.base import InputCodec, register_input_codec
from mlserver.codecs.lists import as_list, is_list_of
from mlserver.types import RequestInput, ResponseOutput, Parameters
from functools import partial


def _pil_base64encode(img: "Image.Image", use_bytes: bool = False) -> Union[bytes, str]:
    buf = io.BytesIO()
    img.save(buf, format=img.format)
    if use_bytes:
        return base64.b64encode(buf.getvalue())
    return base64.b64encode(buf.getvalue()).decode()


def _pil_base64decode(imgbytes: Union[bytes, str]) -> "Image.Image":
    if isinstance(imgbytes, bytes):
        imgbytes = imgbytes.decode()
    buf = io.BytesIO(base64.b64decode(imgbytes))
    return Image.open(buf)


@register_input_codec
class PILImageCodec(InputCodec):
    """
    Codec that convers to / from a PIL.Image input.
    """

    ContentType = "pillow_image"
    TypeHint = List[bytes]

    @classmethod
    def can_encode(cls, payload: Any) -> bool:
        return is_list_of(payload, Image.Image)

    @classmethod
    def encode_output(
        cls, name: str, payload: List[Image.Image], use_bytes: bool = True, **kwargs
    ) -> ResponseOutput:
        packed = map(partial(_pil_base64encode, use_bytes=use_bytes), payload)
        shape = [len(payload), 1]
        return ResponseOutput(
            name=name,
            parameters=Parameters(
                content_type=cls.ContentType,
            ),
            datatype="BYTES",
            shape=shape,
            data=list(packed),
        )

    @classmethod
    def decode_output(cls, response_output: ResponseOutput) -> List[bytes]:
        packed = response_output.data.__root__
        return list(map(_pil_base64decode, as_list(packed)))

    @classmethod
    def encode_input(
        cls, name: str, payload: List[bytes], use_bytes: bool = True, **kwargs
    ) -> RequestInput:
        output = cls.encode_output(name, payload, use_bytes)
        return RequestInput(
            name=output.name,
            parameters=Parameters(
                content_type=cls.ContentType,
            ),
            datatype=output.datatype,
            shape=output.shape,
            data=output.data,
        )

    @classmethod
    def decode_input(cls, request_input: RequestInput) -> List[bytes]:
        packed = request_input.data.__root__
        return list(map(_pil_base64decode, as_list(packed)))



from .base import MultiInputRequestCodec, HuggingfaceRequestCodec
from .image import PILImageCodec
from .json import HuggingfaceSingleJSONCodec
from .jsonlist import HuggingfaceListJSONCodec
from .numpylist import NumpyListCodec
from .conversation import HuggingfaceConversationCodec
from .raw import RawCodec
from .utils import EqualUtil

__all__ = [
    "MultiInputRequestCodec",
    "HuggingfaceRequestCodec",
    "PILImageCodec",
    "HuggingfaceSingleJSONCodec",
    "HuggingfaceListJSONCodec",
    "HuggingfaceConversationCodec",
    "NumpyListCodec",
    "RawCodec",
    "EqualUtil",
]



from typing import Optional, Type, Any, Dict, List, Union, Sequence
from mlserver.codecs.utils import (
    has_decoded,
    _save_decoded,
    get_decoded_or_raw,
)
from mlserver.codecs import StringCodec
from mlserver.codecs.base import (
    RequestCodec,
    register_request_codec,
    InputCodec as InputCodecTy,
)
from mlserver.codecs.base import (
    find_input_codec,
    find_input_codec_by_payload,
)
from mlserver.types import (
    Parameters,
    InferenceRequest,
    InferenceResponse,
    RequestInput,
    ResponseOutput,
)
from .image import PILImageCodec
from .json import HuggingfaceSingleJSONCodec
from .jsonlist import HuggingfaceListJSONCodec
from .conversation import HuggingfaceConversationCodec
from .numpylist import NumpyListCodec
from .raw import RawCodec


def set_content_type(input_v: RequestInput, content_type: str):
    if input_v.parameters is None:
        input_v.parameters = Parameters(content_type=content_type)
        return
    if not input_v.parameters.content_type:
        input_v.parameters.content_type = content_type


class MultiInputRequestCodec(RequestCodec):
    """
    Huggingface codecs is prefered, then mlserver's
    """

    DefaultCodec: Type["InputCodecTy"] = StringCodec
    InputCodecsWithPriority: List[Type[InputCodecTy]] = []
    ContentType = StringCodec.ContentType

    @classmethod
    def _find_encode_codecs(
        cls, payload: Dict[str, Any]
    ) -> Dict[str, Union[Type["InputCodecTy"], "InputCodecTy", None]]:
        field_codec: Dict[str, Union[Type["InputCodecTy"], "InputCodecTy", None]] = {}
        for field, value in payload.items():
            for codec in cls.InputCodecsWithPriority:
                if codec.can_encode(value):
                    field_codec[field] = codec
                    break
            if field not in field_codec:
                field_codec[field] = find_input_codec_by_payload(value)

        return field_codec

    @classmethod
    def _find_decode_codecs(
        cls, data: Union[InferenceResponse, InferenceRequest]
    ) -> Dict[str, Union[Type[InputCodecTy], InputCodecTy, None]]:
        field_codec: Dict[str, Union[Type[InputCodecTy], InputCodecTy, None]] = {}
        default_codec: Union[Type[InputCodecTy], InputCodecTy, None] = None
        fields: Sequence[Union[RequestInput, ResponseOutput]] = []
        if data.parameters and data.parameters.content_type:
            default_codec = find_input_codec(data.parameters.content_type)
        if default_codec is None:
            default_codec = cls.DefaultCodec
        if isinstance(data, InferenceRequest):
            fields = data.inputs
        else:
            fields = data.outputs  # type: ignore
        for field in fields:
            if not field.parameters:
                field_codec[field.name] = default_codec
                continue
            if not field.parameters.content_type:
                field_codec[field.name] = default_codec
                continue
            codec = find_input_codec(field.parameters.content_type)
            if codec:
                field_codec[field.name] = codec
            else:
                field_codec[field.name] = default_codec
        return field_codec

    @classmethod
    def _can_encode_request(cls, payload: Dict[str, Any]) -> bool:
        field_codecs = cls._find_encode_codecs(payload)
        return bool(all(field_codecs.values()))

    @classmethod
    def can_encode(cls, payload: Dict[str, Any]) -> bool:
        """
        Inputs always is Dict, Outputs always is list
        """
        if isinstance(payload, dict):
            return cls._can_encode_request(payload)
        elif isinstance(payload, list):
            return True
        else:
            return False

    @classmethod
    def encode_response(
        cls,
        model_name: str,
        payload: List[Any],
        model_version: Optional[str] = None,
        **kwargs,
    ) -> InferenceResponse:
        """
        Always use HuggingfaceListJSONCodec
        """
        if not isinstance(payload, list):
            payload = [payload]
        return InferenceResponse(
            model_name=model_name,
            model_version=model_version,
            outputs=[HuggingfaceListJSONCodec.encode_output("output", payload)],
        )

    @classmethod
    def decode_response(
        cls, response: InferenceResponse
    ) -> Union[List[Any], Dict[Any, Any]]:
        data = {}
        is_list = True
        field_codecs = cls._find_decode_codecs(response)
        for item in response.outputs:
            if not has_decoded(item):
                codec = field_codecs[item.name]
                if codec is not None:
                    decoded_payload = codec.decode_output(item)
                    _save_decoded(item, decoded_payload)

            value = get_decoded_or_raw(item)
            data[item.name] = value
            if not item.name.startswith("output_"):
                is_list = False
        if not is_list:
            return data
        return [data[key] for key in sorted(data)]

    @classmethod
    def encode_request(cls, payload: Dict[str, Any], **kwargs) -> InferenceRequest:
        field_codecs = cls._find_encode_codecs(payload)
        inputs = []
        for key, value in payload.items():
            codec = field_codecs[key]
            if codec is None:
                raise Exception(
                    f"codec for key {key} value not found, value is {value}"
                )
            input_v = codec.encode_input(key, value, **kwargs)
            set_content_type(input_v, codec.ContentType)
            inputs.append(input_v)
        return InferenceRequest(
            parameters=Parameters(
                content_type=cls.ContentType,
            ),
            inputs=inputs,
        )

    @classmethod
    def decode_request(cls, request: InferenceRequest) -> Dict[str, Any]:
        values = {}
        field_codecs = cls._find_decode_codecs(request)
        for item in request.inputs:
            if not has_decoded(item):
                codec = field_codecs[item.name]
                if codec is not None:
                    decoded_payload = codec.decode_input(item)
                    _save_decoded(item, decoded_payload)

            value = get_decoded_or_raw(item)
            values[item.name] = value
        return values


@register_request_codec
class HuggingfaceRequestCodec(MultiInputRequestCodec):
    InputCodecsWithPriority = [
        PILImageCodec,
        HuggingfaceSingleJSONCodec,
        HuggingfaceListJSONCodec,
        HuggingfaceConversationCodec,
        NumpyListCodec,
        RawCodec,
    ]
    ContentType = StringCodec.ContentType
    DefaultCodec = StringCodec



import io
import json
from typing import Dict, Any, List
import base64
import numpy as np
from PIL import Image, ImageChops
from transformers.pipelines import Conversation

IMAGE_PREFIX = "data:image/"
DEFAULT_IMAGE_FORMAT = "PNG"


class HuggingfaceJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, Image.Image):
            buf = io.BytesIO()
            if not obj.format:
                obj.format = DEFAULT_IMAGE_FORMAT
            obj.save(buf, format=obj.format)
            return (
                IMAGE_PREFIX
                + obj.format.lower()
                + ";base64,"
                + base64.b64encode(buf.getvalue()).decode()
            )
        elif isinstance(obj, Conversation):
            return {
                "uuid": str(obj.uuid),
                "past_user_inputs": obj.past_user_inputs,
                "generated_responses": obj.generated_responses,
                "new_user_input": obj.new_user_input,
            }
        else:
            return json.JSONEncoder.default(self, obj)


def json_encode(payload: Any, use_bytes: bool = False):
    if use_bytes:
        return json.dumps(payload, cls=HuggingfaceJSONEncoder).encode()
    return json.dumps(payload, cls=HuggingfaceJSONEncoder)


def json_decode(payload):
    raw_dict = json.loads(payload)
    return Convertor.do(raw_dict)


conversation_keys = {
    "uuid",
    "past_user_inputs",
    "generated_responses",
    "new_user_input",
}


class Convertor:
    @classmethod
    def do(cls, raw):
        if isinstance(raw, dict):
            return cls.convert_dict(raw)
        elif isinstance(raw, list):
            return cls.convert_list(raw)
        else:
            return raw

    @classmethod
    def convert_conversation(cls, d: Dict[str, Any]):
        if set(d.keys()) == conversation_keys:
            return Conversation(
                text=d["new_user_input"],
                conversation_id=d["uuid"],
                past_user_inputs=d["past_user_inputs"],
                generated_responses=d["generated_responses"],
            )
        return None

    @classmethod
    def convert_dict(cls, d: Dict[str, Any]):
        conversation = cls.convert_conversation(d)
        if conversation is not None:
            return conversation
        tmp = {}
        for k, v in d.items():
            if isinstance(v, dict):
                if set(v.keys()) == conversation_keys:
                    tmp[k] = Conversation(text=v["new_user_input"])
                else:
                    tmp[k] = cls.convert_dict(v)
            elif isinstance(v, list):
                tmp[k] = cls.convert_list(v)
            elif isinstance(v, str):
                if v.startswith(IMAGE_PREFIX):
                    decoded = base64.b64decode(v.split(",")[1])
                    buf = io.BytesIO(decoded)
                    tmp[k] = Image.open(buf)
                else:
                    tmp[k] = v  # type: ignore
            else:
                tmp[k] = v
        return tmp

    @classmethod
    def convert_list(cls, list_data: List[Any]):
        nl = []
        for el in list_data:
            if isinstance(el, list):
                nl.append(cls.convert_list(el))
            elif isinstance(el, dict):
                nl.append(cls.convert_dict(el))
            else:
                nl.append(el)
        return nl


class EqualUtil:
    def pil_equal(img1: "Image.Image", img2: "Image.Image") -> bool:
        diff = ImageChops.difference(img1, img2)
        if diff.getbbox() is None:
            return True
        return False

    @staticmethod
    def list_equal(list1: List[Any], list2: List[Any]) -> bool:
        if len(list1) != len(list2):
            return False
        for idx, el in enumerate(list1):
            if isinstance(el, dict):
                if not EqualUtil.dict_equal(el, list2[idx]):
                    return False
            elif isinstance(el, list):
                if not EqualUtil.list_equal(el, list2[idx]):
                    return False
            elif isinstance(el, Image.Image):
                if not EqualUtil.pil_equal(el, list2[idx]):
                    return False
            elif isinstance(el, np.ndarray):
                if not np.array_equal(el, list2[idx]):
                    return False
            else:
                if el != list2[idx]:
                    return False
        return True

    @staticmethod
    def dict_equal(dict1: Dict[Any, Any], dict2: Dict[Any, Any]) -> bool:
        if not set(dict1.keys()) == set(dict2.keys()):
            return False
        for k, v in dict1.items():
            if isinstance(v, Image.Image):
                pass
            elif isinstance(v, Image.Image):
                if not EqualUtil.pil_equal(v, dict2[k]):
                    return False
            elif isinstance(v, dict):
                if not EqualUtil.dict_equal(v, dict2[k]):
                    return False
            elif isinstance(v, list):
                if not EqualUtil.list_equal(v, dict2[k]):
                    return False
            elif isinstance(v, np.ndarray):
                if not np.array_equal(v, dict2[k]):
                    return False
            else:
                if v != dict2[k]:
                    return False
        return True



import os
import orjson

from typing import Optional, Dict
from pydantic import BaseSettings
from distutils.util import strtobool
from transformers.pipelines import SUPPORTED_TASKS

try:
    # Optimum 1.7 changed the import name from `SUPPORTED_TASKS` to
    # `ORT_SUPPORTED_TASKS`.
    # We'll try to import the more recent one, falling back to the previous
    # import name if not present.
    # https://github.com/huggingface/optimum/blob/987b02e4f6e2a1c9325b364ff764da2e57e89902/optimum/pipelines/__init__.py#L18
    from optimum.pipelines import ORT_SUPPORTED_TASKS as SUPPORTED_OPTIMUM_TASKS
except ImportError:
    from optimum.pipelines import SUPPORTED_TASKS as SUPPORTED_OPTIMUM_TASKS

from mlserver.settings import ModelSettings

from .errors import (
    MissingHuggingFaceSettings,
    InvalidTransformersTask,
    InvalidOptimumTask,
    InvalidModelParameter,
    InvalidModelParameterType,
)

ENV_PREFIX_HUGGINGFACE_SETTINGS = "MLSERVER_MODEL_HUGGINGFACE_"
PARAMETERS_ENV_NAME = "PREDICTIVE_UNIT_PARAMETERS"


class HuggingFaceSettings(BaseSettings):
    """
    Parameters that apply only to HuggingFace models
    """

    class Config:
        env_prefix = ENV_PREFIX_HUGGINGFACE_SETTINGS

    # TODO: Document fields
    task: str = ""
    """
    Pipeline task to load.
    You can see the available Optimum and Transformers tasks available in the
    links below:

    - `Optimum Tasks <https://huggingface.co/docs/optimum/onnxruntime/usage_guides/pipelines#inference-pipelines-with-the-onnx-runtime-accelerator>`_
    - `Transformer Tasks <https://huggingface.co/docs/transformers/task_summary>`_
    """  # noqa: E501

    task_suffix: str = ""
    """
    Suffix to append to the base task name.
    Useful for, e.g. translation tasks which require a suffix on the task name
    to specify source and target.
    """

    pretrained_model: Optional[str] = None
    """
    Name of the model that should be loaded in the pipeline.
    """

    pretrained_tokenizer: Optional[str] = None
    """
    Name of the tokenizer that should be loaded in the pipeline.
    """

    framework: Optional[str] = None
    """
    The framework to use, either "pt" for PyTorch or "tf" for TensorFlow.
    """

    optimum_model: bool = False
    """
    Flag to decide whether the pipeline should use a Optimum-optimised model or
    the standard Transformers model.
    Under the hood, this will enable the model to use the optimised ONNX
    runtime.
    """

    device: int = -1
    """
    Device in which this pipeline will be loaded (e.g., "cpu", "cuda:1", "mps",
    or a GPU ordinal rank like 1).
    """

    inter_op_threads: Optional[int] = None
    """
    Threads used for parallelism between independent operations.
    PyTorch:
    https://pytorch.org/docs/stable/notes/cpu_threading_torchscript_inference.html
    Tensorflow:
    https://www.tensorflow.org/api_docs/python/tf/config/threading/set_inter_op_parallelism_threads
    """

    intra_op_threads: Optional[int] = None
    """
    Threads used within an individual op for parallelism.
    PyTorch:
    https://pytorch.org/docs/stable/notes/cpu_threading_torchscript_inference.html
    Tensorflow:
    https://www.tensorflow.org/api_docs/python/tf/config/threading/set_intra_op_parallelism_threads
    """

    @property
    def task_name(self):
        if self.task == "translation":
            return f"{self.task}{self.task_suffix}"
        return self.task


def parse_parameters_from_env() -> Dict:
    """
    This method parses the environment variables injected via SCv1.
    """
    # TODO: Once support for SCv1 is deprecated, we should remove this method and rely
    # purely on settings coming via the `model-settings.json` file.
    parameters = orjson.loads(os.environ.get(PARAMETERS_ENV_NAME, "[]"))

    type_dict = {
        "INT": int,
        "FLOAT": float,
        "DOUBLE": float,
        "STRING": str,
        "BOOL": bool,
    }

    parsed_parameters = {}
    for param in parameters:
        name = param.get("name")
        value = param.get("value")
        type_ = param.get("type")
        if type_ == "BOOL":
            parsed_parameters[name] = bool(strtobool(value))
        else:
            try:
                parsed_parameters[name] = type_dict[type_](value)
            except ValueError:
                raise InvalidModelParameter(name, value, type_)
            except KeyError:
                raise InvalidModelParameterType(type_)
    return parsed_parameters


def get_huggingface_settings(model_settings: ModelSettings) -> HuggingFaceSettings:
    env_params = parse_parameters_from_env()
    if not env_params and (
        not model_settings.parameters or not model_settings.parameters.extra
    ):
        raise MissingHuggingFaceSettings()

    extra = env_params or model_settings.parameters.extra  # type: ignore
    hf_settings = HuggingFaceSettings(**extra)  # type: ignore

    if hf_settings.task not in SUPPORTED_TASKS:
        raise InvalidTransformersTask(hf_settings.task, SUPPORTED_TASKS.keys())

    if hf_settings.optimum_model:
        if hf_settings.task not in SUPPORTED_OPTIMUM_TASKS:
            raise InvalidOptimumTask(hf_settings.task, SUPPORTED_OPTIMUM_TASKS.keys())

    return hf_settings



from typing import List

from mlserver.errors import MLServerError


class MissingHuggingFaceSettings(MLServerError):
    def __init__(self):
        super().__init__("Missing HuggingFace Runtime settings.")


class InvalidTransformersTask(MLServerError):
    def __init__(self, task: str, available_tasks: List[str]):
        msg = f"Invalid transformer task: {task}. Available tasks: {available_tasks}."
        super().__init__(msg)


class InvalidOptimumTask(MLServerError):
    def __init__(self, task: str, available_tasks: List[str]):
        msg = (
            "Invalid transformer task for Optimum model: {task}. "
            f"Available Optimum tasks: {available_tasks}."
        )
        super().__init__(msg)


class InvalidModelParameter(MLServerError):
    def __init__(self, name: str, value: str, param_type: str):
        msg = (
            f"Bad model parameter: {name}"
            f" with value {value}"
            f" can't be parsed as a {param_type}"
        )
        super().__init__(msg)


class InvalidModelParameterType(MLServerError):
    def __init__(self, param_type: str):
        msg = (
            f"Bad model parameter type: {param_type}."
            f" Only valid types are INT, FLOAT, DOUBLE, STRING, BOOL."
        )
        super().__init__(msg)
